//
//  main.c
//  CGImageCarbon
//
//  Created by David Phillip Oster on 3/27/07.
// Based on a newsgroup question: wrap a 2D array of 16-bit per pixel pixels as an image
//

#include <Carbon/Carbon.h>

static OSStatus        AppEventHandler( EventHandlerCallRef inCaller, EventRef inEvent, void* inRefcon );
static OSStatus        HandleNew();
static OSStatus        WindowEventHandler( EventHandlerCallRef inCaller, EventRef inEvent, void* inRefcon );

static IBNibRef        sNibRef;
static CGImageRef      sCGImage;
static CGDataProviderRef sDataProvider;

static int sWidth;
static int sHeight;
static char *sMax;
static unsigned short *sImageArray;
static char *sP;

static size_t DataProviderGetBytes(void *info, void *buffer, size_t count){
  if (sMax < sP) {
    count = 0;
  }else if (sMax < sP+count){
    count = sMax - sP;
  }
  memmove(buffer, sP, count);
  sP += count;
  return count;
}

static void DataProviderSkipBytes(void *info, size_t count){
  sP += count;
}

static void DataProviderRewind(void *info){
  sP = (char *) sImageArray;
}

static void DataProviderReleaseInfo(void *info){
}

static void InitImage(CGRect frame){
  if (sImageArray){
    return;
  }
  sWidth = frame.size.width;
  sWidth = ((sWidth + (16-1)) / 16) * 16;
  sHeight = frame.size.height;
  // initialize our 2 d array of pixels.
  sImageArray = (unsigned short *) malloc(sWidth*sHeight *sizeof(unsigned short));
  sMax = (char *) (sImageArray+sWidth*sHeight);
  sP = (char *) sImageArray;
  int i, j, k;
  for(j = k = 0; j < sHeight;  ++j) {
    unsigned short *p = &sImageArray[j*sWidth];
    for(i = 0; i < sWidth; ++i, ++k, ++p) {
#if 0
      *p = ((int)(i * (255./sWidth)) & 0xFF) |  // alpha
        ((int)(i * (255./sWidth)) << 8) | // blue
        ((int)(i * (255./sWidth)) << 16) |  // green
        ( 0xFF) << 24;  // red
#else
      *p = ((int)(i * (31./sWidth)) & 0x1F) << 1 |
        ((int)(i * (31./sWidth)) & 0x1F) << 6 |
        ((int)(i * (31./sWidth)) & 0x1F) << 11;
#endif
    }
  }
  static CGDataProviderCallbacks callbacks = {
   DataProviderGetBytes,
   DataProviderSkipBytes,
   DataProviderRewind,
   DataProviderReleaseInfo
  };
  sDataProvider = CGDataProviderCreate(NULL, &callbacks);
  CGColorSpaceRef colorspace = CGColorSpaceCreateDeviceRGB();
  sCGImage =  CGImageCreate(sWidth, sHeight, 5, 16, sWidth*2, colorspace, 
      kCGBitmapByteOrder32Host, sDataProvider, NULL, FALSE, kCGRenderingIntentDefault);
  CGColorSpaceRelease(colorspace);
}

//--------------------------------------------------------------------------------------------
int main(int argc, char* argv[]){
  OSStatus                    err;
  static const EventTypeSpec    kAppEvents[] = {
      { kEventClassCommand, kEventCommandProcess }
  };

  // Create a Nib reference, passing the name of the nib file (without the .nib extension).
  // CreateNibReference only searches into the application bundle.
  err = CreateNibReference( CFSTR("main"), &sNibRef );
  require_noerr( err, CantGetNibRef );
  
  // Once the nib reference is created, set the menu bar. "MainMenu" is the name of the menu bar
  // object. This name is set in InterfaceBuilder when the nib is created.
  err = SetMenuBarFromNib( sNibRef, CFSTR("MenuBar") );
  require_noerr( err, CantSetMenuBar );
  
  // Install our handler for common commands on the application target
  InstallApplicationEventHandler( NewEventHandlerUPP( AppEventHandler ),
                                  GetEventTypeCount( kAppEvents ), kAppEvents,
                                  0, NULL );
  
  // Create a new window. A full-fledged application would do this from an AppleEvent handler
  // for kAEOpenApplication.
  HandleNew();
  
  // Run the event loop
  RunApplicationEventLoop();

CantSetMenuBar:
CantGetNibRef:
  return err;
}

//--------------------------------------------------------------------------------------------
static OSStatus AppEventHandler( EventHandlerCallRef inCaller, EventRef inEvent, void* inRefcon ){
  OSStatus    result = eventNotHandledErr;

  if (kEventClassCommand == GetEventClass( inEvent ) ){
      HICommandExtended cmd;
      verify_noerr( GetEventParameter( inEvent, kEventParamDirectObject, typeHICommand, NULL, sizeof( cmd ), NULL, &cmd ) );
      
      if(kEventCommandProcess == GetEventKind( inEvent ) ){
          switch ( cmd.commandID ){
              case kHICommandNew:
                  result = HandleNew();
                  break;
                  
              // Add your own command-handling cases here
              
              default:
                  break;
          }
      }          
  }
  return result;
}

DEFINE_ONE_SHOT_HANDLER_GETTER( WindowEventHandler )

//--------------------------------------------------------------------------------------------
static void DrawUserPane(ControlRef pane, SInt16 part){
  Rect         rect;
  CGContextRef cg = NULL;
  GetControlBounds(pane, &rect);
  WindowRef window = GetControlOwner(pane);
  CGrafPtr winPort = GetWindowPort(window);
  QDBeginCGContext(winPort, &cg);
  CGRect cgRect = CGRectMake(rect.left, rect.top, rect.right-rect.left, rect.bottom-rect.top);

  CGContextDrawImage(cg, cgRect, sCGImage);
  QDEndCGContext(winPort, &cg);
}

//--------------------------------------------------------------------------------------------
static OSStatus HandleNew(void){
  OSStatus                    e = noErr;
  WindowRef                   window;
  ControlRef                  pane;
  static const ControlID      kPaneID = {0, 100};
  static const EventTypeSpec  kWindowEvents[] =
  {
      { kEventClassCommand, kEventCommandProcess }
  };
  static ControlUserPaneDrawUPP drawupp = NULL;
  
  // Create a window. "MainWindow" is the name of the window object. This name is set in 
  // InterfaceBuilder when the nib is created.
  if(noErr == e){ e = CreateWindowFromNib( sNibRef, CFSTR("MainWindow"), &window ); }
  if(noErr == e){
    // Install a command handler on the window. We don't use this handler yet, but nearly all
    // Carbon apps will need to handle commands, so this saves everyone a little typing.
    e = InstallWindowEventHandler( window, GetWindowEventHandlerUPP(),
                               GetEventTypeCount( kWindowEvents ), kWindowEvents,
                               window, NULL );
  }
  // Position new windows in a staggered arrangement on the main screen
  if(noErr == e){ e = RepositionWindow( window, NULL, kWindowCascadeOnMainScreen ); }
  if(noErr == e){ e = GetControlByID(window, &kPaneID, &pane); }
  if(noErr == e){
    Rect         rect;
    GetControlBounds(pane, &rect);
    CGRect cgRect = CGRectMake(rect.left, rect.top, rect.right-rect.left, rect.bottom-rect.top);
    InitImage(cgRect);
  }
  if(NULL == drawupp){ drawupp = NewControlUserPaneDrawUPP(DrawUserPane); }
  if(noErr == e){
    e = SetControlData(pane, kControlEntireControl, kControlUserPaneDrawProcTag, sizeof drawupp, &drawupp); 
  }
  
  // The window was created hidden, so show it
  ShowWindow( window );
  
  return e;
}

//--------------------------------------------------------------------------------------------
static OSStatus WindowEventHandler( EventHandlerCallRef inCaller, EventRef inEvent, void* inRefcon ){
  OSStatus    err = eventNotHandledErr;
  
  if(kEventClassCommand == GetEventClass(inEvent)){
    HICommandExtended cmd;
    verify_noerr( GetEventParameter( inEvent, kEventParamDirectObject, typeHICommand, NULL, sizeof( cmd ), NULL, &cmd ) );
    
    if(kEventCommandProcess == GetEventKind(inEvent)){
      switch ( cmd.commandID ){
          // Add your own command-handling cases here
          
      default:
          break;
      }
    }
  }
  
  return err;
}

// Just draw a red rect. I used this while debugging DrawUserPane
#if 0
  float components[] = { 1., 0., 0., 1.};
  CGColorSpaceRef space = CGColorSpaceCreateDeviceRGB();
  CGColorRef color = CGColorCreate(space, components);
  CGContextSetFillColorWithColor(cg, color);
  CGContextFillRect(cg, cgRect);
  CGColorRelease(color);
  CGColorSpaceRelease(space);
#endif
